STARTUP(cellular_credentials_set("hologram", "", "", NULL));

#include <JsonParserGeneratorRK.h>
#include <math.h>

const String ID = "002";

const int COND_PIN = A4;  // Pin assignments
const int THERMISTOR_PIN = A5;
const int LED_PIN = D0;
const int POWER_1 = D1;
const int POWER_2 = D2;
const int POWER_3 = D3;
const int POWER_4 = D4;
const int TRIG_PIN = D5;
const int ECHO_PIN = D6;

SYSTEM_THREAD(ENABLED);  // Allows the device to take sample measurements while attempting to make a cloud connection
SYSTEM_MODE(SEMI_AUTOMATIC);

ApplicationWatchdog wd(300000, shutdown);  // Turns off the device if measurement is not taken successfully after 5 minutes

FuelGauge fuel;

int runtime;
int sleepTime = 60;
bool updateAvailable = false;
bool firstRun = true;

void setup() {
    Particle.subscribe("hook-response/liveConfig", getLiveConf, MY_DEVICES);
    Particle.connect();
    RGB.control(true);
    RGB.brightness(3);
    RGB.control(false);

    pinMode(TRIG_PIN, OUTPUT);
    pinMode(ECHO_PIN, INPUT);

    pinMode(POWER_1, OUTPUT);
    pinMode(POWER_2, OUTPUT);
    pinMode(POWER_3, OUTPUT);
    pinMode(POWER_4, OUTPUT);
    pinMode(LED_PIN, OUTPUT);
}

void loop() {
    runtime = millis();

    if (firstRun) {
        digitalWrite(LED_PIN, HIGH);
        delay(250);
        digitalWrite(LED_PIN, LOW);
    }

    digitalWrite(POWER_1, HIGH);  // Turns on switched power
    digitalWrite(POWER_2, HIGH);
    digitalWrite(POWER_3, HIGH);
    digitalWrite(POWER_4, HIGH);
    
    delay(500);
    int range = getRange();
    JsonWriterStatic<256> doc;  // Assembles sensor data into JSON string to be uploaded
    {
        JsonWriterAutoObject obj(&doc);

        doc.insertKeyValue("id", ID);
        doc.insertKeyValue("battery", fuel.getVCell());  // Gets sensor data
        doc.insertKeyValue("stage", range);
        doc.insertKeyValue("temp", getTemp());
        doc.insertKeyValue("conductivity", getCond());
    }
    delay(500);
    digitalWrite(POWER_1, LOW);  // Turns off switched power
    digitalWrite(POWER_2, LOW);
    digitalWrite(POWER_3, LOW);
    digitalWrite(POWER_4, LOW);

    while (!Particle.connected()) {  // Waits here until device establishes a cloud connection
        if (firstRun) {
            delay(900);
            digitalWrite(LED_PIN, HIGH);
            delay(200);
            digitalWrite(LED_PIN, LOW);
        } else {
            delay(1000);
        }
    }

    Particle.publish("liveConfig", PRIVATE);
    delay(4000);

    if (updateAvailable) {
        digitalWrite(LED_PIN, HIGH);
        Particle.publish("ready_for_update");
        delay(120000);
    }

    Particle.publish("Measurement", doc.getBuffer());  // Publishes sensor measurement and sleeps
    shutdown();
}

void getLiveConf(const char *event, const char *data) {
    JsonParserStatic<1024, 10> parser;
    parser.addString(data);
    if (parser.parse()) {
        sleepTime = parser.getReference().key(ID).key("update_freq").valueInt();
        updateAvailable = parser.getReference().key(ID).key("ota_update").valueBool();
    }
}

void shutdown() {  // Runs when a measurement is taken, or if the device fails to connect in time

    for (int i = 0; i < 7; i++) {
        if (firstRun) {
            digitalWrite(LED_PIN, HIGH);
            delay(50);
            digitalWrite(LED_PIN, LOW);
            delay(50);
        } else {
            delay(100);
        }
    }
    firstRun = false;
    runtime = millis() - runtime;
    System.sleep(SLEEP_MODE_SOFTPOWEROFF, (sleepTime * 60) - (runtime / 1000) - 6);
}

/*
int recursionMax = 5;   // Maximum number of times the device will try to get a stable stage measurement before giving up
int getRange() {
    if(recursionMax >= 0){
        recursionMax--;
        int sum = 0;
        int lastVal = -1;
        bool badReading = false;
    
        for (int i = 0; i < 5; i++) {  // Takes 5 ultrasonic measurements and returns the average in milliseconds
            digitalWrite(TRIG_PIN, LOW);
            delayMicroseconds(2);
            digitalWrite(TRIG_PIN, HIGH);
            delayMicroseconds(10);
            digitalWrite(TRIG_PIN, LOW);
            int measurement = rdPulseIn(ECHO_PIN, HIGH, 37) / 5.82;
            
        if (lastVal != -1 && abs(measurement - lastVal) > 100){
                badReading = true;
            }
            
            sum += measurement;
            lastVal = measurement;
            delay(250);
        }
        
        if (badReading){
            return getRange();
        }
        else {
            recursionMax = 5;
            return sum / 5;
        }
    }
    
    else{
        recursionMax = 5;
        return -1;
    }
   
}


void bubbleSort(int arr[], int n) {
    int i, j;
    for (i = 0; i < n-1; i++) {
        for (j = 0; j < n-i-1; j++) {
            if (arr[j] > arr[j+1]) {
                swap(&arr[j], &arr[j+1]);  
            }
        }
    }
}

void swap(int *xp, int *yp) {
    int temp = *xp;
    *xp = *yp;
    *yp = temp;
}
*/
int getRange() {
    int maxVal = -1;

    for (int i = 0; i < 5; i++) {  // Takes 5 ultrasonic measurements and returns the average in milliseconds
        digitalWrite(TRIG_PIN, LOW);
        delayMicroseconds(2);
        digitalWrite(TRIG_PIN, HIGH);
        delayMicroseconds(10);
        digitalWrite(TRIG_PIN, LOW);
        int measurement = rdPulseIn(ECHO_PIN, HIGH, 37) / 5.82;
        
        if (measurement > maxVal) {
            maxVal = measurement;
        }
        delay(250);
    }
    
    return maxVal;
}


double getCond() {
    double sum = 0;

    for (int i = 0; i < 5; i++) {
        sum += map(analogRead(COND_PIN), 0.0, 4095.0, 0.0, 1.0);  // Take 5 conductivity measurements and return average as value from 0 to 1
    }
    return sum / 5;
}

double getTemp() {
    const int SAMPLE_NUMBER = 5;  // Constants for thermistor readings
    const double BALANCE_RESISTOR = 97200.0;
    const double BETA = 3974.0;
    const double ROOM_TEMP = 298.15;
    const double RESISTOR_ROOM_TEMP = 100000.0;
    const double MAX_ADC = 4095.0;

    double rThermistor = 0;
    double tKelvin = 0;
    double tCelsius = 0;
    double adcAverage = 0;
    int adcSamples[SAMPLE_NUMBER];

    for (int i = 0; i < SAMPLE_NUMBER; i++) {
        adcSamples[i] = analogRead(THERMISTOR_PIN);
        delay(10);
    }

    for (int i = 0; i < SAMPLE_NUMBER; i++) {
        adcAverage += adcSamples[i];  // Take the average of 5 temperature readings
    }
    adcAverage /= SAMPLE_NUMBER;

    rThermistor = BALANCE_RESISTOR * ((MAX_ADC / adcAverage) - 1);
    tKelvin = (BETA * ROOM_TEMP) /
              (BETA + (ROOM_TEMP * log(rThermistor / RESISTOR_ROOM_TEMP)));
    tCelsius = tKelvin - 273.15;
    return tCelsius;  // Return the temperature in Celsius
}

unsigned long rdPulseIn(int pin, int value, int timeout) {  // Counts microseconds between sending and receieving a signal from the ultrasonic sensor

    unsigned long now = micros();
    while (pinReadFast(pin) == value) {
        if (micros() - now > (timeout * 1000)) {
            return 0;
        }
    }

    now = micros();
    while (pinReadFast(pin) != value) {
        if (micros() - now > (timeout * 1000)) {
            return 0;
        }
    }

    now = micros();
    while (pinReadFast(pin) == value) {
        if (micros() - now > (timeout * 1000)) {
            return 0;
        }
    }
    return micros() - now;
}
